#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

"""
Implements the base class for `TextRank` –
with placeholder methods to be used by the subclasses for algorithm extensions.
"""

from collections import Counter, defaultdict, OrderedDict
from dataclasses import dataclass
import json
import math
import pathlib
import time
import typing

from icecream import ic  # type: ignore # pylint: disable=E0401,W0611 # lgtm[py/unused-import]
from spacy.tokens import Doc, Span, Token  # type: ignore # pylint: disable=E0401
import graphviz  # type: ignore # pylint: disable=E0401
import networkx as nx  # type: ignore # pylint: disable=E0401

from .util import groupby_apply, default_scrubber

try:
    import altair as alt  # type: ignore # pylint: disable=E0401
    import pandas as pd  # type: ignore # pylint: disable=E0401
except ImportError:
    _has_altair_and_pandas = False
else:
    _has_altair_and_pandas = True

# parameter type annotation for a *stop words* source
StopWordsLike = typing.Union[ str, pathlib.Path, typing.Dict[str, typing.List[str]] ]


@dataclass(order=True, frozen=True)
class Lemma:
    """
A data class representing one node in the *lemma graph*.
    """
    lemma: str
    pos: str


    def label (
        self
        ) -> str:
        """
Generates a more simplified string representation than `repr()`
provides.

    returns:
string representation
        """
        return str((self.lemma, self.pos,))


@dataclass
class Phrase:
    """
A data class representing one ranked phrase.
    """
    text: str
    chunks: typing.List[Span]
    count: int
    rank: float


@dataclass
class Sentence:
    """
A data class representing the distance measure for one sentence.
    """
    start: int
    end: int
    sent_id: int
    phrases: typing.Set[int]
    distance: float


    def empty (
        self
        ) -> bool:
        """
Test whether this sentence includes any ranked phrases.

    returns:
`True` if the `phrases` is not empty.
        """
        return len(self.phrases) == 0


    def text (
        self,
        doc: Doc,
        ) -> str:
        """
Accessor for the text slice of the `spaCy` [`Doc`](https://spacy.io/api/doc)
document represented by this sentence.

    doc:
source document

    returns:
the sentence text
        """
        return doc[self.start:self.end]


@dataclass
class VectorElem:
    """
A data class representing one element in the *unit vector* of the document.
    """
    phrase: Phrase
    phrase_id: int
    coord: float


class BaseTextRankFactory:
    """
A factory class that provides the document with its instance of
`BaseTextRank`
    """

    _EDGE_WEIGHT: float = 1.0
    _POS_KEPT: typing.List[str] = ["ADJ", "NOUN", "PROPN", "VERB"]
    _TOKEN_LOOKBACK: int = 3


    def __init__ (
        self,
        *,
        edge_weight: float = _EDGE_WEIGHT,
        pos_kept: typing.List[str] = None,
        token_lookback: int = _TOKEN_LOOKBACK,
        scrubber: typing.Optional[typing.Callable] = None,
        stopwords: typing.Optional[StopWordsLike] = None,
        ) -> None:
        """
Constructor for a factory used to instantiate the PyTextRank pipeline components.

    edge_weight:
default weight for an edge

    pos_kept:
parts of speech tags to be kept; adjust this if strings representing the POS tags change

    token_lookback:
the window for neighboring tokens – similar to a *skip gram*

    scrubber:
optional "scrubber" function to clean up punctuation from a token; if `None` then defaults to `pytextrank.default_scrubber`; when running, PyTextRank will throw a `FutureWarning` warning if the configuration uses a deprecated approach for a scrubber function

    stopwords:
optional dictionary of `lemma: [pos]` items to define the *stop words*, where each item has a key as a lemmatized token and a value as a list of POS tags; may be a file name (string) or a [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html) for a JSON file; otherwise throws a `TypeError` exception
        """
        self.edge_weight: float = edge_weight
        self.token_lookback: int = token_lookback

        if pos_kept:
            self.pos_kept: typing.List[str] = pos_kept
        else:
            self.pos_kept = self._POS_KEPT

        if scrubber:
            self.scrubber: typing.Callable = scrubber
        else:
            self.scrubber = default_scrubber

        if stopwords:
            self.stopwords: typing.Dict[str, typing.List[str]] = self._load_stopwords(stopwords)
        else:
            self.stopwords = defaultdict(list)


    @classmethod
    def _load_stopwords (
        cls,
        stopwords: typing.Optional[StopWordsLike] = None,
        ) -> typing.Dict[str, typing.List[str]]:
        """
Load a dictionary of
[*stop words*](https://derwen.ai/docs/ptr/glossary/#stop-words)
– i.e., tokens to be ignored when constructing the
[*lemma graph*](https://derwen.ai/docs/ptr/glossary/#lemma-graph).

Note: be cautious about the use of this feature, since it can get
"greedy" and bias or otherwise distort the results.

    stopwords:
optional dictionary of `lemma: [pos]` items to define the *stop words*, where each item has a key as a lemmatized token and a value as a list of POS tags; may be a file name (string) or a [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html) for a JSON file; otherwise throws a `TypeError` exception

    returns:
the *stop words* dictionary
        """
        if isinstance(stopwords, dict):
            return stopwords

        if isinstance(stopwords, pathlib.Path):
            path: pathlib.Path = stopwords
        else:
            path = pathlib.Path(str)  # type: ignore

        if path.exists():
            with open(path, "r") as f:
                data = json.load(f)

                if data:
                    return data.items()

        raise TypeError("cannot parse the stopwords source as a dictionary")


    def __call__ (
        self,
        doc: Doc,
        ) -> Doc:
        """
Set the extension attributes on a `spaCy` [`Doc`](https://spacy.io/api/doc)
document to create a *pipeline component* for `BaseTextRank` as
a stateful component, invoked when the document gets processed.

See: <https://spacy.io/usage/processing-pipelines#pipelines>

    doc:
a document container, providing the annotations produced by earlier stages of the `spaCy` pipeline
        """
        Doc.set_extension("textrank", force=True, default=None)
        Doc.set_extension("phrases", force=True, default=[])

        doc._.textrank = BaseTextRank(
            doc,
            edge_weight = self.edge_weight,
            pos_kept = self.pos_kept,
            token_lookback = self.token_lookback,
            scrubber = self.scrubber,
            stopwords = self.stopwords,
            )

        doc._.phrases = doc._.textrank.calc_textrank()
        return doc


class BaseTextRank:
    """
Implements the *TextRank* algorithm defined by
[[mihalcea04textrank]](https://derwen.ai/docs/ptr/biblio/#mihalcea04textrank),
deployed as a `spaCy` pipeline component.

This class does not get called directly; instantiate its factory
instead.
    """

    def __init__ (
        self,
        doc: Doc,
        edge_weight: float,
        pos_kept: typing.List[str],
        token_lookback: int,
        scrubber: typing.Callable,
        stopwords: typing.Dict[str, typing.List[str]],
        ) -> None:
        """
Constructor for a `TextRank` object.

    doc:
a document container, providing the annotations produced by earlier stages of the `spaCy` pipeline

    edge_weight:
default weight for an edge

    pos_kept:
parts of speech tags to be kept; adjust this if strings representing the POS tags change

    token_lookback:
the window for neighboring tokens – similar to a *skip gram*

    scrubber:
optional "scrubber" function to clean up punctuation from a token

    stopwords:
optional dictionary of `lemma: [pos]` items to define the *stop words*, where each item has a key as a lemmatized token and a value as a list of POS tags
        """
        self.doc: Doc = doc
        self.edge_weight: float = edge_weight
        self.token_lookback: int = token_lookback
        self.pos_kept: typing.List[str] = pos_kept
        self.scrubber: typing.Callable = scrubber
        self.stopwords: typing.Dict[str, typing.List[str]] = stopwords

        # internal data for BiasedTextRank
        self.focus_tokens: typing.Set[str] = set()
        self.node_bias = 1.0
        self.default_bias = 1.0

        # effectively, performs the same work as the `reset()` method;
        # called explicitly here for the sake of type annotations
        self.elapsed_time: float = 0.0
        self.lemma_graph: nx.DiGraph = nx.DiGraph()
        self.phrases: typing.List[Phrase] = []
        self.ranks: typing.Dict[Lemma, float] = {}
        self.seen_lemma: typing.Dict[Lemma, typing.Set[int]] = OrderedDict()


    def reset (
        self
        ) -> None:
        """
Reinitialize the data structures needed for extracting phrases,
removing any pre-existing state.
        """
        self.elapsed_time = 0.0
        self.lemma_graph = nx.DiGraph()
        self.phrases = []
        self.ranks = {}
        self.seen_lemma = OrderedDict()


    def calc_textrank (
        self
        ) -> typing.List[Phrase]:
        """
Iterate through each sentence in the doc, constructing a
[*lemma graph*](https://derwen.ai/docs/ptr/glossary/#lemma-graph)
then returning the top-ranked phrases.

This method represents the heart of the *TextRank* algorithm.

    returns:
list of ranked phrases, in descending order
        """
        t0 = time.time()
        self.reset()
        self.lemma_graph = self._construct_graph()

        # to run the algorithm, we use the NetworkX implementation
        # for PageRank (i.e., based on eigenvector centrality)
        # to calculate a rank for each node in the lemma graph
        self.ranks = nx.pagerank(
            self.lemma_graph,
            personalization = self.get_personalization(),
            )

        # agglomerate the lemmas ranked in the lemma graph into ranked
        # phrases, leveraging information from earlier stages of the
        # pipeline: noun chunks and named entities
        nc_phrases: typing.Dict[Span, float] = self._collect_phrases(self.doc.noun_chunks, self.ranks)
        ent_phrases: typing.Dict[Span, float] = self._collect_phrases(self.doc.ents, self.ranks)
        all_phrases: typing.Dict[Span, float] = { **nc_phrases, **ent_phrases }

        # since noun chunks can be expressed in different ways (e.g., may
        # have articles or prepositions), we need to find a minimum span
        # for each phrase based on combinations of lemmas
        raw_phrase_list: typing.List[Phrase] = self._get_min_phrases(all_phrases)
        phrase_list: typing.List[Phrase] = sorted(raw_phrase_list, key=lambda p: p.rank, reverse=True)

        t1 = time.time()
        self.elapsed_time = (t1 - t0) * 1000.0

        return phrase_list


    def get_personalization (  # pylint: disable=R0201
        self
        ) -> typing.Optional[typing.Dict[Lemma, float]]:
        """
Get the *node weights* for initializing the use of the
[*Personalized PageRank*](https://derwen.ai/docs/ptr/glossary/#personalized-pagerank)
algorithm.

Defaults to a no-op for the base *TextRank* algorithm.

    returns:
`None`
        """
        return None


    def _construct_graph (
        self
        ) -> nx.DiGraph:
        """
Construct the
[*lemma graph*](https://derwen.ai/docs/ptr/glossary/#lemma-graph).

    returns:
a directed graph representing the lemma graph
        """
        g = nx.DiGraph()

        # add nodes made of Lemma(lemma, pos)
        g.add_nodes_from(self.node_list)

        # add edges between nodes that co-occur within a window,
        # weighted by the count
        g.add_edges_from(self.edge_list)

        return g


    def _keep_token (
        self,
        token: Token,
        ) -> bool:
        """
Skip tokens that are either in the stop word list or don't have a part
of speech tag suitable for vertices in the lemma graph.
Otherwise, track this token in the `seen_lemma` dictionary.

    token:
a parsed `spaCy` [`Token`](https://spacy.io/api/token) to be evaluated

    returns:
boolean value for whether to keep this token as a node in the lemma graph
        """
        lemma = token.lemma_.lower().strip()

        if self._is_stopword(lemma, token):
            return False

        if token.pos_ not in self.pos_kept:
            return False

        # also track occurrence of this token's lemma, for later use
        key = Lemma(lemma, token.pos_,)

        if key not in self.seen_lemma:
            self.seen_lemma[key] = set([token.i])
        else:
            self.seen_lemma[key].add(token.i)

        return True


    def _is_stopword (
        self,
        lemma: str,
        token: Token,
        ) -> bool:
        """
Determine whether the given (lemma, pos) pair was identified with in
the *stop words* list.

    returns:
boolean result of the test
        """
        return lemma in self.stopwords and token.pos_ in self.stopwords[lemma]


    @property
    def node_list (
        self
        ) -> typing.List[Lemma]:
        """
Build a list of vertices for the lemma graph.

    returns:
list of nodes
        """
        nodes: typing.List[Lemma] = [
            Lemma(token.lemma_, token.pos_)
            for token in self.doc
            if self._keep_token(token)
        ]

        return nodes


    @property
    def edge_list (
        self
        ) -> typing.List[typing.Tuple[Lemma, Lemma, typing.Dict[str, float]]]:
        """
Build a list of weighted edges for the lemma graph.

    returns:
list of weighted edges
        """
        edges: typing.List[typing.Tuple[Lemma, Lemma]] = []

        for sent in self.doc.sents:
            h = [
                Lemma(token.lemma_, token.pos_)
                for token in sent
                if self._keep_token(token)
            ]

            for hop in range(self.token_lookback):
                for idx, node in enumerate(h[: -1 - hop]):
                    nbor = h[hop + idx + 1]
                    edges.append((node, nbor))

        # include weight on the edge: (2, 3, {'weight': 3.1415})
        weighted_edges: typing.List[typing.Tuple[Lemma, Lemma, typing.Dict[str, float]]] = [
            (*n, {"weight": w * self.edge_weight}) for n, w in Counter(edges).items()
        ]

        return weighted_edges


    def _collect_phrases (
        self,
        spans: typing.Iterable[Span],
        ranks: typing.Dict[Lemma, float]
        ) -> typing.Dict[Span, float]:
        """
Aggregate the rank metrics of the individual nodes (tokens) within
each phrase.

    spans:
spans of noun chunks

    ranks:
rank metrics corresponding to each node

    returns:
phrases extracted from the lemma graph, each with an aggregate rank metric
        """
        phrases: typing.Dict[Span, float] = {
            span: sum(
                ranks[Lemma(token.lemma_, token.pos_)]
                for token in span
                if self._keep_token(token)
            )
            for span in spans
        }

        return {
            span: self._calc_discounted_normalised_rank(span, sum_rank)
            for span, sum_rank in phrases.items()
        }


    def _calc_discounted_normalised_rank (
        self,
        span: Span,
        sum_rank: float
        ) -> float:
        """
Since the noun chunking is greedy, we discount the ranks using a point
estimate based on the number of non-lemma tokens within a phrase.

    span:
a span representing one phrase

    sum_rank:
sum of the ranks for each token within this span

    returns:
normalized rank metric
        """
        non_lemma = len([tok for tok in span if tok.pos_ not in self.pos_kept])
        non_lemma_discount = len(span) / (len(span) + (2.0 * non_lemma) + 1.0)

        # NB:
        # This implements a *point estimate* for the ratio of the span
        # of an extracted phrase divided by the number of non-lemma
        # tokens within that span.
        #
        # The algorithm defined as in the original paper
        # [[mihalcea04textrank]](https://derwen.ai/docs/ptr/biblio/#mihalcea04textrank)
        # only considered multi-word phrases which had adjacent lemmas.
        # Early use cases in industry (circa 2009) required better recall
        # of phrases so this approach of using a point estimate here to
        # "normalize" the rank metric causes the longer phrases to get
        # discounted when they included too many non-lemma tokens.
        # In other words, we allow some non-lemmas, but avoid having
        # the algorithm become too "greedy" as it builds phrases.
        #
        # Kudos to @debraj135 who asked for an explanation about this
        # section of the code.
        phrase_rank = math.sqrt(sum_rank / (len(span) + non_lemma))

        return phrase_rank * non_lemma_discount


    def _get_min_phrases (
        self,
        all_phrases: typing.Dict[Span, float]
        ) -> typing.List[Phrase]:
        """
Group the phrases by their text content, selecting the span with the
maximum rank within each group, then collect the ranked phrases into
an ordered list.

Throws a `FutureWarning` warning if the configuration uses a
deprecated approach for a scrubber function

    all_phrases:
the raw phrase list

    returns:
an ordered list of ranked phrases

        """
        try:
            data: typing.List[typing.Tuple[Span, float, Span]] = [
                (self.scrubber(span), rank, span) for span, rank in all_phrases.items()
            ]
        except AttributeError:
            raise FutureWarning("Text-based scrubbers are deprecated. Use a `Span` instead.")

        keyfunc = lambda x: x[0]
        applyfunc = lambda g: list((rank, spans) for text, rank, spans in g)

        phrases: typing.List[typing.Tuple[str, typing.List[typing.Tuple[float, Span]]]] = groupby_apply(
            data,
            keyfunc,
            applyfunc,
        )

        phrase_list: typing.List[Phrase] = [
            Phrase(
                text = p[0],
                rank = max(rank for rank, span in p[1]),
                count = len(p[1]),
                chunks = list(span for rank, span in p[1]),
            )
            for p in phrases
        ]

        return phrase_list


    def get_unit_vector (
        self,
        limit_phrases: int,
        ) -> typing.List[VectorElem]:
        """
Construct a *unit vector* representing the top-ranked phrases in a
`spaCy` [`Doc`](https://spacy.io/api/doc) document.
This provides a *characteristic* for comparing each sentence to the
entire document.
Taking the ranked phrases in descending order, the unit vector is a
normalized list of their calculated ranks, up to the specified limit.

    limit_phrases:
maximum number of top-ranked phrases to use in the *unit vector*

    returns:
the unit vector, as a list of `VectorElem` objects
        """
        unit_vector: typing.List[VectorElem] = [
            VectorElem(
                phrase = p,
                phrase_id = phrase_id,
                coord = p.rank,
                )
            for phrase_id, p in enumerate(self.doc._.phrases)
            ]

        # truncate to the specified limit
        limit = min(limit_phrases, len(unit_vector))
        unit_vector = unit_vector[:limit]

        # normalize the phrase coordinates, such that the unit vector
        # has length = 1.0
        sum_length = sum([ elem.coord for elem in unit_vector ])

        for elem in unit_vector:
            if sum_length > 0.0:
                elem.coord = elem.coord / sum_length
            else:
                elem.coord = 0.0

        return unit_vector


    def calc_sent_dist (
        self,
        limit_phrases: int,
        ) -> typing.List[Sentence]:
        """
For each sentence in the document, calculate its distance from a *unit
vector* of top-ranked phrases.

    limit_phrases:
maximum number of top-ranked phrases to use in the *unit vector*

    returns:
a list of sentence distance measures
        """
        unit_vector = self.get_unit_vector(limit_phrases)

        sent_dist: typing.List[Sentence] = [
            Sentence(
                start = s.start,
                end = s.end,
                sent_id = sent_id,
                phrases = set(),
                distance = 0.0,
                )
            for sent_id, s in enumerate(self.doc.sents)
            ]

        # identify the top-ranked phrases in each sentence
        for elem in unit_vector:
            for chunk in elem.phrase.chunks:
                for sent in sent_dist:
                    if chunk.start >= sent.start and chunk.start <= sent.end:
                        sent.phrases.add(elem.phrase_id)
                        break

        # calculate a euclidean distance for each sentence; in other
        # words, test for inclusion of each phrase in the unit vector
        for sent in sent_dist:
            sum_sq = 0.0

            for elem in unit_vector:
                if elem.phrase_id not in sent.phrases:
                    sum_sq += elem.coord**2.0

            sent.distance = math.sqrt(sum_sq)

        return sent_dist


    def summary (
        self,
        *,
        limit_phrases: int = 10,
        limit_sentences: int = 4,
        preserve_order: bool = False,
        ) -> typing.Iterator[str]:
        """
Run an
[*extractive summarization*](https://derwen.ai/docs/ptr/glossary/#extractive-summarization),
based on the vector distance (per sentence) for each of the top-ranked phrases.

    limit_phrases:
maximum number of top-ranked phrases to use in the distance vectors

    limit_sentences:
total number of sentences to yield for the extractive summarization

    preserve_order:
flag to preserve the order of sentences as they originally occurred in the source text; defaults to `False`

    yields:
texts for sentences, in order
        """
        # build a list of sentence indices sorted by distance
        sent_dist: typing.List[Sentence] = self.calc_sent_dist(limit_phrases)

        top_sent_ids: typing.List[int] = [
            sent.sent_id
            for sent in sorted(sent_dist, key=lambda sent: sent.distance)
            ]

        # truncated to the specified limit
        limit = min(limit_sentences, len(top_sent_ids))
        top_sent_ids = top_sent_ids[:limit]

        # optional: sort in ascending order of index to preserve the
        # order in which sentences appear in the original text
        if preserve_order:
            top_sent_ids.sort()

        # extract sentences with the least distance, up to the limit
        # requested
        for sent_id in top_sent_ids:
            yield sent_dist[sent_id].text(self.doc)


    def write_dot (
        self,
        *,
        path: typing.Optional[typing.Union[ str, pathlib.Path ]] = "graph.dot",
        ) -> None:
        """
Serialize the lemma graph in the `Dot` file format.

    path:
path for the output file; defaults to `"graph.dot"`
        """
        dot = graphviz.Digraph()

        for lemma in self.lemma_graph.nodes():
            rank = self.ranks[lemma]
            key = lemma.label()

            label = "{} ({:.4f})".format(key, rank)
            dot.node(key, label)

        for edge in self.lemma_graph.edges():
            dot.edge(edge[0].label(), edge[1].label(), constraint="false")

        if isinstance(path, str):
            path = pathlib.Path(path)

        with open(path, "w") as f:  # type: ignore
            f.write(dot.source)


    def plot_keyphrases (
        self
        ) -> typing.Any:
        """
Plot a document's keyphrases rank profile using
[`altair.Chart`](https://altair-viz.github.io/user_guide/generated/toplevel/altair.Chart.html)

Throws an `ImportError` if the `altair` and `pandas` libraries are not installed.

    returns:
the `altair` chart being rendered
        """
        if not _has_altair_and_pandas:
            raise ImportError("altair and pandas are required to use this method. Install them with `pip install 'pytextrank[viz]'`")

        source = pd.DataFrame([p.__dict__ for p in self.doc._.phrases]).drop("chunks", axis=1).reset_index()

        c = (
            alt.Chart(source)
            .mark_bar()
            .encode(x="index", y="rank", color="count", tooltip=["text", "rank", "count"])
            .properties(title="Keyphrase profile of the document")
        )

        return c



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore

"""Shared fixture functions."""
import pathlib
import pytest  # pylint: disable=E0401
import spacy  # pylint: disable=E0401
from spacy.language import Language  # pylint: disable=E0401
from spacy.tokens import Doc  # pylint: disable=E0401


@pytest.fixture(scope="session")
def nlp () -> Language:
    """
Language shared fixture.
    """
    nlp = spacy.load("en_core_web_sm")  # pylint: disable=W0621
    return nlp


@pytest.fixture(scope="session")
def doc (nlp: Language) -> Doc:  # pylint: disable=W0621
    """
Doc shared fixture.

    returns:
spaCy EN doc containing a piece of football news.
    """
    text = pathlib.Path("dat/cfc.txt").read_text()
    doc = nlp(text)  # pylint: disable=W0621
    return doc


@pytest.fixture(scope="session")
def long_doc (nlp: Language) -> Doc:  # pylint: disable=W0621
    """
Doc shared fixture.

    returns:
spaCy EN doc containing a long text.
    """
    text = pathlib.Path("dat/lee.txt").read_text()
    doc = nlp(text)  # pylint: disable=W0621
    return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore

"""Unit tests for BiasedRank."""
from spacy.tokens import Doc  # pylint: disable=E0401

import sys
sys.path.insert(0, "../pytextrank")

from pytextrank.base import BaseTextRankFactory  # pylint: disable=E0401
from pytextrank.biasedrank import BiasedTextRankFactory  # pylint: disable=E0401


def test_default_biased_rank (doc: Doc):
    """
Biasedrank should behave like base textrank by default.
    """
    # given
    biased_rank = BiasedTextRankFactory()
    base_text_rank = BaseTextRankFactory()

    # when
    processed_doc = base_text_rank(doc)
    phrases = processed_doc._.phrases

    comparison_doc = biased_rank(doc)
    comparison_phrases = comparison_doc._.phrases

    # then
    assert tuple(p.text for p in phrases) == tuple(p.text for p in comparison_phrases)


def test_biased_rank (long_doc: Doc):
    """
Rank phrases close to 'focus' higher.
    """
    # given
    biased_rank = BiasedTextRankFactory()
    base_text_rank = BaseTextRankFactory()

    # when
    processed_doc = base_text_rank(long_doc)
    phrases = processed_doc._.phrases

    # the article is primarily about the match up between AlphaGo and
    # Grandmaster Lee Sedol
    assert "grandmaster Lee Sedol" in [p.text for p in phrases][:3]
    assert "Gary Kasparov" not in [p.text for p in phrases][:3]

    comparison_doc = biased_rank(long_doc)
    tr = comparison_doc._.textrank
    tr.change_focus(
        "Chess",
        bias=10.0,
        default_bias=0.0)

    biased_phrases = comparison_doc._.phrases

    # then
    # shifting the focus to chess bring Gary Kasparov in top ranks
    assert "Gary Kasparov" in [p.text for p in biased_phrases][:3]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

import pathlib
import typing

from spacy.language import Language  # type: ignore # pylint: disable=E0401

from .base import BaseTextRankFactory, BaseTextRank, Lemma, Phrase, Sentence, VectorElem, StopWordsLike
from .biasedrank import BiasedTextRankFactory, BiasedTextRank
from .positionrank import PositionRankFactory, PositionRank
from .util import groupby_apply, default_scrubber, maniacal_scrubber, split_grafs, filter_quotes
from .version import MIN_PY_VERSION, _versify, _check_version, __version__


######################################################################
## add component factories to the spaCy pipeline namespace

_DEFAULT_CONFIG = {
    "edge_weight": BaseTextRankFactory._EDGE_WEIGHT,  # pylint: disable=W0212
    "pos_kept": BaseTextRankFactory._POS_KEPT,  # pylint: disable=W0212
    "token_lookback": BaseTextRankFactory._TOKEN_LOOKBACK,  # pylint: disable=W0212
    "scrubber": None,
    "stopwords": None,
    }


# wrap these factory definitions in a try-catch, so they don't cause
# exceptions during programmatic loading with `importlib`

try:
    @Language.factory("textrank", default_config=_DEFAULT_CONFIG)
    def _create_component_tr (
        nlp: Language,  # pylint: disable=W0613
        name: str,  # pylint: disable=W0613
        edge_weight: float,
        pos_kept: typing.List[str],
        token_lookback: int,
        scrubber: typing.Optional[typing.Callable],
        stopwords: typing.Optional[StopWordsLike],
        ) -> BaseTextRankFactory:
        """
Component factory for the `TextRank` base class.
        """
        return BaseTextRankFactory(
            edge_weight = edge_weight,
            pos_kept = pos_kept,
            token_lookback = token_lookback,
            scrubber = scrubber,
            stopwords = stopwords,
        )


    @Language.factory("positionrank", default_config=_DEFAULT_CONFIG)
    def _create_component_pr (
        nlp: Language,  # pylint: disable=W0613
        name: str,  # pylint: disable=W0613
        edge_weight: float,
        pos_kept: typing.List[str],
        token_lookback: int,
        scrubber: typing.Optional[typing.Callable],
        stopwords: typing.Optional[StopWordsLike],
        ) -> PositionRankFactory:
        """
Component factory for the `PositionRank` extended class.
        """
        return PositionRankFactory(
            edge_weight = edge_weight,
            pos_kept = pos_kept,
            token_lookback = token_lookback,
            scrubber = scrubber,
            stopwords = stopwords,
        )


    @Language.factory("biasedtextrank", default_config=_DEFAULT_CONFIG)
    def _create_component_br (
        nlp: Language,  # pylint: disable=W0613
        name: str,  # pylint: disable=W0613
        edge_weight: float,
        pos_kept: typing.List[str],
        token_lookback: int,
        scrubber: typing.Optional[typing.Callable],
        stopwords: typing.Optional[StopWordsLike],
        ) -> BiasedTextRankFactory:
        """
Component factory for the `BiasedTextRank` extended class.
        """
        return BiasedTextRankFactory(
            edge_weight = edge_weight,
            pos_kept = pos_kept,
            token_lookback = token_lookback,
            scrubber = scrubber,
            stopwords = stopwords,
        )

except Exception:  # pylint: disable=W0703
    print("ERROR: it appears that `spaCy` 3.x has not been imported?")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore

"""Unit tests for PositionRank."""
from spacy.tokens import Doc  # pylint: disable=E0401

import sys
sys.path.insert(0, "../pytextrank")

from pytextrank.base import BaseTextRankFactory  # pylint: disable=E0401
from pytextrank.positionrank import PositionRankFactory  # pylint: disable=E0401


def test_position_rank (doc: Doc):
    """
Ranks keywords that appear early in the document higher than TextRank
does.
    """
    # given
    position_rank = PositionRankFactory()
    base_text_rank = BaseTextRankFactory()

    # when
    processed_doc = position_rank(doc)
    phrases = processed_doc._.phrases
    comparison_doc = base_text_rank(doc)
    comparison_phrases = comparison_doc._.phrases

    # then
    assert set(p.rank for p in phrases) != set(p.rank for p in comparison_phrases)

    # The test article mentions `Chelsea` at the beginning of the
    # article while it mentions `Shanghai Shenhua` anecdotally later
    # in the article. However, with normal TextRank, `Shanghai
    # Shenhua` is part of top 10 phrases and `Chelsea` isn't. With
    # PositionRank, the situation is the opposite, which is the
    # desired outcome when parsing a news article.

    assert "Chelsea" in [p.text for p in phrases[:10]]
    assert "Chelsea" not in [p.text for p in comparison_phrases[:10]]
    assert "Shanghai Shenhua" not in [p.text for p in phrases[:10]]
    assert "Shanghai Shenhua" in [p.text for p in comparison_phrases[:10]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

import sys
import typing


######################################################################
## Python version checking

MIN_PY_VERSION: typing.Tuple = (3, 7,)
__version__: str = "3.2.0"


def _versify (
    py_version_info: typing.Tuple
    ) -> str:
    """
Semiprivate helper function to convert Python version to a point release (a string).

    py_version_info:
Python version info as a named tuple from the operating system, e.g., from
[`sys.version_info[:2]`](https://docs.python.org/3/library/sys.html#sys.version_info)

    returns:
Python version info in [*semantic versioning*](https://semver.org/) format
    """
    return ".".join([ str(x) for x in py_version_info ])


def _check_version () -> None:
    """
Semiprivate helper function to check the Python version info versus
the minimum required for **pytextrank**.

Throws a `RuntimeError` if the installed Python interpreter is out of
date.
    """
    py_version_info: typing.Tuple = sys.version_info[:2]

    if py_version_info < MIN_PY_VERSION:
        error_msg = "This version of pytextrank requires Python {} or later ({} detected)\n"
        raise RuntimeError(error_msg.format(_versify(MIN_PY_VERSION), _versify(py_version_info)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore

"""Unit tests for BaseTextRank."""
from spacy.language import Language  # pylint: disable=E0401
from spacy.tokens import Span, Doc  # pylint: disable=E0401
import spacy  # pylint: disable=E0401

import sys
sys.path.insert(0, "../pytextrank")

from pytextrank.base import BaseTextRankFactory  # pylint: disable=E0401


def test_base_text_rank (doc: Doc):
    """
Ranks unique keywords in a document correctly, sorted decreasing by
centrality.
    """
    # given
    base_text_rank = BaseTextRankFactory()

    # when
    processed_doc = base_text_rank(doc)
    phrases = processed_doc._.phrases

    # then
    assert len(phrases) > 0
    assert len(set(p.text for p in phrases)) == len(phrases)
    assert phrases[0].rank == max(p.rank for p in phrases)


def test_add_pipe (nlp: Language):
    """
Works as a pipeline component and can be disabled.
    """
    # given
    # base_text_rank = BaseTextRankFactory()
    nlp.add_pipe("textrank", last=True)

    # works as a pipeline component
    # when
    text = "linear constraints over the"
    doc = nlp(text)
    phrases = [ p.text for p in doc._.phrases ]

    # then
    assert len(doc._.phrases) > 0
    assert any(map(lambda x: "constraints" in x, phrases))

    # identifies phrases not in noun chunks
    # when
    text = """
everything you need to know about student loan interest rates variable
and fixed rates capitalization amortization student loan refinancing
and more.
"""

    doc = nlp(text)
    phrases = [ p.text for p in doc._.phrases ]

    # then
    assert len(doc._.phrases) >= 2

    # resolves Py 3.5 dict KeyError
    # when
    text = "linear constraints over the set of natural numbers"
    doc = nlp(text)
    phrases = [ p.text for p in doc._.phrases ]

    # then
    assert any(map(lambda x: "constraints" in x, phrases))

    # pipeline can be disabled
    # when
    with nlp.select_pipes(disable=["textrank"]):
        doc = nlp(text)

        # then
        assert len(doc._.phrases) == 0


def test_summary (nlp: Language):
    """
Summarization produces the expected results.
Limit evaluation to Top-K
    """
    # given
    LIMIT_PHRASES = 10
    TOP_K = 5

    expected_trace = [
        [0, [0, 9, 2, 6]],
        [1, [9]],
        [2, [2]],
        [3, [7]],
        [4, [8]],
    ]

    with open("dat/lee.txt", "r") as f:
        text = f.read()
        doc = nlp(text)
        tr = doc._.textrank

        # calculates *unit vector* and sentence distance measures
        # when
        trace = [
            [ sent_dist.sent_id, list(sent_dist.phrases) ]
            for sent_dist in tr.calc_sent_dist(limit_phrases=LIMIT_PHRASES)
            if not sent_dist.empty()
            ]

        # then
        assert trace[:TOP_K] == expected_trace


def test_multiple_summary (nlp: Language):
    """
Summarization produces consistent results when called across multiple
docs.
    """
    texts = []

    with open("dat/lee.txt", "r") as f:
        text = f.read()
        texts.append(text)

    with open("dat/mih.txt", "r") as f:
        text = f.read()
        texts.append(text)

    docs = [nlp(text) for text in texts]

    trace1 = [
        [sent_dist.sent_id, list(sent_dist.phrases)]
        for sent_dist in docs[0]._.textrank.calc_sent_dist(limit_phrases=10)
        if not sent_dist.empty()
    ]

    trace2 = [
        [sent_dist.sent_id, list(sent_dist.phrases)]
        for sent_dist in docs[1]._.textrank.calc_sent_dist(limit_phrases=10)
        if not sent_dist.empty()
    ]

    assert trace1 != trace2


def test_stop_words ():
    """
Works as a pipeline component and can be disabled.
    """
    nlp1 = spacy.load("en_core_web_sm")
    nlp1.add_pipe("textrank")

    # "words" is in top phrases
    with open("dat/gen.txt", "r") as f:
        doc = nlp1(f.read())

        phrases = [
            phrase.text
            for phrase in doc._.phrases[:5]
            ]

        assert "words" in phrases

    # add `"word": ["NOUN"]` to the *stop words*, to remove instances
    # of `"word"` or `"words"` then see how the ranked phrases differ?

    nlp2 = spacy.load("en_core_web_sm")
    nlp2.add_pipe("textrank", config={ "stopwords": { "word": ["NOUN"] } })

    with open("dat/gen.txt", "r") as f:
        doc = nlp2(f.read())

        phrases = [
            phrase.text
            for phrase in doc._.phrases[:5]
            ]

        assert "words" not in phrases


def test_scrubber ():
    """
Works as a pipeline component and can be disabled.
    """
    # given

    text = "This is a test for scrubber."

    nlp1 = spacy.load("en_core_web_sm")
    nlp1.add_pipe("textrank", last=True)

    doc = nlp1(text)

    phrases = [
        phrase.text
        for phrase in doc._.phrases
    ]

    assert "a test" in phrases

    @spacy.registry.misc("articles_scrubber")
    def articles_scrubber():  # pylint: disable=W0612
        def scrubber_func(text_span: Span) -> str:
            for token in text_span:
                if token.pos_ != "DET":
                    break
                text_span = text_span[1:]
            return text_span.text

        return scrubber_func


    # add "articles_scrubber" to config
    # then see if phrases have `articles` in it?

    nlp2 = spacy.load("en_core_web_sm")
    nlp2.add_pipe("textrank", config={"stopwords": {"test": ["NOUN"]}, "scrubber": {"@misc": "articles_scrubber"}}, last=True)


    # then
    doc = nlp2(text)

    phrases = [
        phrase.text
        for phrase in doc._.phrases
    ]

    assert "test" in phrases and "a test" not in phrases



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

"""
Implements the *Biased TextRank* algorithm.
"""

import typing

from spacy.tokens import Doc, Token  # type: ignore # pylint: disable=E0401

from .base import BaseTextRankFactory, BaseTextRank, Lemma, Phrase


class BiasedTextRankFactory (BaseTextRankFactory):
    """
A factory class that provides the document with its instance of
`BiasedTextRank`
    """

    def __call__ (
        self,
        doc: Doc,
        ) -> Doc:
        """
Set the extension attributes on a `spaCy` [`Doc`](https://spacy.io/api/doc)
document to create a *pipeline component* for `BiasedTextRank` as
a stateful component, invoked when the document gets processed.

See: <https://spacy.io/usage/processing-pipelines#pipelines>

    doc:
a document container, providing the annotations produced by earlier stages of the `spaCy` pipeline
        """
        Doc.set_extension("textrank", force=True, default=None)
        Doc.set_extension("phrases", force=True, default=[])

        doc._.textrank = BiasedTextRank(
            doc,
            edge_weight = self.edge_weight,
            pos_kept = self.pos_kept,
            token_lookback = self.token_lookback,
            scrubber = self.scrubber,
            stopwords = self.stopwords,
            )

        doc._.phrases = doc._.textrank.calc_textrank()
        return doc


class BiasedTextRank (BaseTextRank):
    """
Implements the *Biased TextRank* algorithm described by
[[kazemi-etal-2020-biased]](https://derwen.ai/docs/ptr/biblio/#kazemi-etal-2020-biased),
deployed as a `spaCy` pipeline component.

This class does not get called directly; instantiate its factory
instead.
    """

    _DEFAULT_BIAS: float = 1.0


    def _get_node_bias (
        self,
        token: Token,
        ) -> float:
        """
Assign the bias to use, where focus nodes get the preset bias and
other nodes get the default (`1.0`) value.

    token:
`spaCy` token to use for lookup in the *focus set*

    returns:
bias to apply for the *node weight*
        """
        if token.text in self.focus_tokens:
            return self.node_bias

        if token.lemma_ in self.focus_tokens:
            return self.node_bias

        return self.default_bias


    def get_personalization (
        self,
        ) -> typing.Optional[typing.Dict[Lemma, float]]:
        """
Get the *node weights* for initializing the use of the
[*Personalized PageRank*](https://derwen.ai/docs/ptr/glossary/#personalized-pagerank)
algorithm.

    returns:
biased restart probabilities to use in the *PageRank* algorithm.
        """
        # TODO: # pylint: disable=W0511
        # => lookup bias based on (lemma, pos) instead, similar to *PositionRank*?

        weighted_nodes: typing.Dict[Lemma, float] = {
            Lemma(token.lemma_, token.pos_): self._get_node_bias(token)
            for token in self.doc
            if token.pos_ in self.pos_kept
        }

        # normalize weights
        total_weight = sum(weighted_nodes.values())

        if total_weight == 0.0:
            return None

        normalized_weighted_nodes: typing.Dict[Lemma, float] = {
            lemma : weight / total_weight
            for lemma, weight in weighted_nodes.items()
            }

        return normalized_weighted_nodes


    def change_focus (
        self,
        focus: str = None,
        bias: float = _DEFAULT_BIAS,
        default_bias: float = _DEFAULT_BIAS,
        ) -> typing.List[Phrase]:
        """
Re-runs the *Biased TextRank* algorithm with the given focus.
This approach allows an application to "change focus" without
re-running the entire pipeline.

    focus:
optional text (string) with space-delimited tokens to use for the *focus set*; defaults to `None`

    bias:
optional bias for *node weight* values on tokens found within the *focus set*; defaults to `1.0`

    default_bias:
optional bias for *node weight* values on tokens not found within the *focus set*; set to `0.0` to enhance the focus, especially in the case of long documents; defaults to `1.0`

    returns:
list of ranked phrases, in descending order
        """
        # update the focus parameters
        if focus:
            self.focus_tokens = set(focus.lower().split(" "))
        else:
            self.focus_tokens = set()

        self.node_bias = bias
        self.default_bias = default_bias

        # update the textrank phrase extraction
        self.doc._.phrases = self.calc_textrank()

        return self.doc._.phrases



# type: ignore

import importlib.util
import pathlib
import setuptools
import typing


KEYWORDS = [
    "biased textrank",
    "entity linking",
    "extractive summarization",
    "graph algorithms",
    "knowledge graph",
    "natural language processing",
    "nlp",
    "parsing",
    "phrase extraction",
    "pipeline component",
    "positionrank",
    "spacy",
    "text analytics",
    "textgraphs",
    "textrank",
    ]


def parse_requirements_file (filename: str) -> typing.List:
    """read and parse a Python `requirements.txt` file, returning as a list of str"""
    with pathlib.Path(filename).open() as f:
        return [ l.strip().replace(" ", "") for l in f.readlines() ]


if __name__ == "__main__":
    spec = importlib.util.spec_from_file_location("pytextrank.version", "pytextrank/version.py")
    pytr_version = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(pytr_version)
    pytr_version._check_version()  # pylint: disable=W0212

    base_packages = parse_requirements_file("requirements.txt")
    docs_packages = parse_requirements_file("requirements-dev.txt")
    viz_packages = parse_requirements_file("requirements-viz.txt")

    setuptools.setup(
        name="pytextrank",
        version = pytr_version.__version__,

        python_requires = ">=" + pytr_version._versify(pytr_version.MIN_PY_VERSION),  # pylint: disable=W0212
        packages = setuptools.find_packages(exclude=[ "docs", "examples" ]),
        install_requires = base_packages,
        extras_require = {
            "base": base_packages,
            "docs": docs_packages,
            "viz": viz_packages
            },

        author="Paco Nathan",
        author_email="paco@derwen.ai",
        license="MIT",

        description="Python implementation of TextRank as a spaCy pipeline extension, for graph-based natural language work plus related knowledge graph practices; used for for phrase extraction and lightweight extractive summarization of text documents.",
        long_description = pathlib.Path("README.md").read_text(),
        long_description_content_type = "text/markdown",

        keywords = ", ".join(KEYWORDS),
        classifiers = [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Information Technology",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Human Machine Interfaces",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Text Processing :: General",
            "Topic :: Text Processing :: Indexing",
            "Topic :: Text Processing :: Linguistic",
            ],

        url = "https://derwen.ai/docs/ptr/",
        project_urls = {
            "Source": "http://github.com/DerwenAI/pytextrank",
            "spaCy uniVerse": "https://spacy.io/universe/project/spacy-pytextrank",
            "Issue Tracker": "https://github.com/DerwenAI/pytextrank/issues",
            "Discussion Forum": "https://www.linkedin.com/groups/6725785/",
            "StackOverflow": "https://stackoverflow.com/search?q=pytextrank",
            "Citations": "https://scholar.google.com/scholar?q=related:5tl6J4xZlCIJ:scholar.google.com/&scioq=&hl=en&as_sdt=0,5",
            },

        zip_safe=False,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

"""
Implements the *PositionRank* algorithm.
"""

import typing

from spacy.tokens import Doc  # type: ignore # pylint: disable=E0401

from .base import BaseTextRankFactory, BaseTextRank, Lemma
from .util import groupby_apply


class PositionRankFactory (BaseTextRankFactory):
    """
A factory class that provides the document with its instance of
`PositionRank`
    """


    def __call__ (
        self,
        doc: Doc,
        ) -> Doc:
        """
Set the extension attributes on a `spaCy` [`Doc`](https://spacy.io/api/doc)
document to create a *pipeline component* for `PositionRank` as
a stateful component, invoked when the document gets processed.

See: <https://spacy.io/usage/processing-pipelines#pipelines>

    doc:
a document container, providing the annotations produced by earlier stages of the `spaCy` pipeline
        """
        Doc.set_extension("textrank", force=True, default=None)
        Doc.set_extension("phrases", force=True, default=[])

        doc._.textrank = PositionRank(
            doc,
            edge_weight = self.edge_weight,
            pos_kept = self.pos_kept,
            token_lookback = self.token_lookback,
            scrubber = self.scrubber,
            stopwords = self.stopwords,
            )

        doc._.phrases = doc._.textrank.calc_textrank()
        return doc


class PositionRank (BaseTextRank):
    """
Implements the *PositionRank* algorithm described by
[[florescuc17]](https://derwen.ai/docs/ptr/biblio/#florescuc17),
deployed as a `spaCy` pipeline component.

This class does not get called directly; instantiate its factory
instead.
    """

    def get_personalization (
        self
        ) -> typing.Optional[typing.Dict[Lemma, float]]:
        """
Get the *node weights* for initializing the use of the
[*Personalized PageRank*](https://derwen.ai/docs/ptr/glossary/#personalized-pagerank)
algorithm.

From the cited reference:

> Specifically, we propose to assign a higher probability to a word
> found on the 2nd position as compared with a word found on the 50th
> position in the same document. The weight of each candidate word is
> equal to its inverse position in the document.  If the same word
> appears multiple times in the target document, then we sum all its
> position weights.

> For example, a word v_i occurring in the following positions: 2nd,
> 5th and 10th, has a weight p(v_i) = 1/2 + 1/5 + 1/10 = 4/5 = 0.8
> The weights of words are normalized before they are used in the
> position-biased PageRank.

    returns:
Biased restart probabilities to use in the *PageRank* algorithm.
        """
        weighted_tokens: typing.List[typing.Tuple[str, float]] = [
            (tok, 1 / (i + 1))
            for i, tok in enumerate(
                token.lemma_ for token in self.doc if token.pos_ in self.pos_kept
            )
        ]

        keyfunc = lambda x: x[0]
        applyfunc = lambda g: sum(w for text, w in g)

        accumulated_weighted_tokens: typing.List[typing.Tuple[str, float]] = groupby_apply(
            weighted_tokens,
            keyfunc,
            applyfunc,
        )

        accumulated_weighted_tokens = sorted(
            accumulated_weighted_tokens, key=lambda x: x[1]
        )

        norm_weighted_tokens = {
            k: w / sum(w_ for _, w_ in accumulated_weighted_tokens)
            for k, w in accumulated_weighted_tokens
        }

        # while the authors assign higher probability to a "word",
        # our *lemma graph* vertices are (lemma, pos) tuples,
        # therefore we map each `Lemma` weight to all the *lemma
        # graph* vertices which contain it

        # TODO: # pylint: disable=W0511
        # => should this map to (lemma, pos) pairs instead?

        weighted_nodes: typing.Dict[Lemma, float] = {
            Lemma(token.lemma_, token.pos_): norm_weighted_tokens[token.lemma_]
            for token in self.doc
            if token.pos_ in self.pos_kept
        }

        return weighted_nodes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/pytextrank#license-and-copyright

import itertools
import re
import string
import typing
import unicodedata

from spacy.tokens import Span  # type: ignore # pylint: disable=E0401


def groupby_apply (
    data: typing.Iterable[typing.Any],
    keyfunc: typing.Callable,
    applyfunc: typing.Callable,
    ) -> typing.List[typing.Tuple[typing.Any, typing.Any]]:
    """
GroupBy using a key function and an apply function, without a `pandas`
dependency.
See: <https://docs.python.org/3/library/itertools.html#itertools.groupby>

    data:
iterable

    keyfunc:
callable to define the key by which you want to group

    applyfunc:
callable to apply to the group

    returns:
an iterable with the accumulated values
    """
    data = sorted(data, key=keyfunc)

    accum: typing.List[typing.Tuple[typing.Any, typing.Any]] = [
        (k, applyfunc(g),)
        for k, g in itertools.groupby(data, keyfunc)
        ]

    return accum


######################################################################
## utility functions

def default_scrubber (
    span: Span
    ) -> str:
    """
Removes spurious punctuation from the given text.
Note: this is intended for documents in English.

    span:
input text `Span`

    returns:
scrubbed text
    """
    return span.text.replace("'", "")


def maniacal_scrubber (
    span: Span
    ) -> str:
    """
Applies multiple approaches for aggressively removing garbled Unicode
and spurious punctuation from the given text.

OH: "It scrubs the garble from its stream... or it gets the debugger again!"

    span:
input text `Span`

    returns:
scrubbed text
    """
    # some web content returns "not string" ??
    # ostensibly this is no longer possible in Py 3.x ...
    # even so some crazy-making "mixed modes" of character encodings
    # have been found in the wild -- YMMV
    text = span.text

    if type(text).__name__ != "str":
        print("not a string?", type(text), text)

    x = " ".join(map(lambda s: s.strip(), text.split("\n"))).strip()

    x = x.replace('“', '"').replace('”', '"')
    x = x.replace("‘", "'").replace("’", "'").replace("`", "'")
    x = x.replace("…", "...").replace("–", "-")

    x = str(unicodedata.normalize("NFKD", x).encode("ascii", "ignore").decode("utf-8"))

    return x


def split_grafs (
    lines: typing.List[str]
    ) -> typing.Iterator[str]:
    """
Segments a raw text, given as a list of lines, into paragraphs.

    lines:
the raw text document, split into a lists of lines

    yields:
text per paragraph
    """
    graf: typing.List[str] = []

    for line in lines:
        line = line.strip()

        if len(line) < 1:
            if len(graf) > 0:
                yield "\n".join(graf)
                graf = []
        else:
            graf.append(line)

    if len(graf) > 0:
        yield "\n".join(graf)


def filter_quotes (
    text: str,
    *,
    is_email: bool = True,
    ) -> typing.List[str]:
    """
Filter the quoted text out of an email message.
This handles quoting methods for popular email systems.

    text:
raw text data

    is_email:
flag for whether the text comes from an email message;
defaults to `True`

    returns:
the filtered text representing as a list of lines
    """
    _PAT_FORWARD = re.compile(r"\n-+ Forwarded message -+\n")
    _PAT_REPLIED = re.compile(r"\nOn.*\d+.*\n?wrote\:\n+\>")
    _PAT_UNSUBSC = re.compile(r"\n-+\nTo unsubscribe,.*\nFor additional commands,.*")

    if is_email:
        text = str(filter(lambda x: x in string.printable, text))

        # strip off quoted text in a forward
        m = _PAT_FORWARD.split(text, re.M)

        if m and len(m) > 1:
            text = m[0]

        # strip off quoted text in a reply
        m = _PAT_REPLIED.split(text, re.M)

        if m and len(m) > 1:
            text = m[0]

        # strip off any trailing unsubscription notice
        m = _PAT_UNSUBSC.split(text, re.M)

        if m:
            text = m[0]

    # replace any remaining quoted text with blank lines
    lines: typing.List[str] = []

    for line in text.split("\n"):
        if line.startswith(">"):
            lines.append("")
        else:
            lines.append(line)

    return list(split_grafs(lines))
